import pytest
import os

from unittest.mock import patch
from mqtt_middlware_utils import logging_utils as utils


class TestLoggingUtils:

    test_log_file = "tests/files/test_log.log"

    def clear_log_file(self):
        if os.path.exists(self.test_log_file):
            os.remove(self.test_log_file)

    def read_file_line(self):
        if os.path.exists(self.test_log_file):
            with open(self.test_log_file, "r") as f:
                return f.readline()

    def test_map_log_level_debug(self):
        log_level_string = "DEBUG"
        mapped_log_level = utils.map_log_level(log_level_string)
        assert mapped_log_level == utils.logging.DEBUG

    def test_map_log_level_info(self):
        log_level_string = "INFO"
        mapped_log_level = utils.map_log_level(log_level_string)
        assert mapped_log_level == utils.logging.INFO

    def test_map_log_level_warning(self):
        log_level_string = "WARNING"
        mapped_log_level = utils.map_log_level(log_level_string)
        assert mapped_log_level == utils.logging.WARNING

    def test_map_log_level_error(self):
        log_level_string = "ERROR"
        mapped_log_level = utils.map_log_level(log_level_string)
        assert mapped_log_level == utils.logging.ERROR

    def test_map_log_level_undefined_log_level(self):
        log_level_string = "WRONG_LEVEL"
        mapped_log_level = utils.map_log_level(log_level_string)
        assert mapped_log_level == utils.logging.INFO

    def test_init_logger(self):
        log_level = "WARNING"
        logger_name = "test_logger"
        log_destination = "tests/files/test_log.log"
        initialized_logger = utils.init_logger(
            log_level, logger_name, log_destination)

        assert initialized_logger.level == utils.logging.WARNING
        assert initialized_logger.name == logger_name
        assert len(initialized_logger.handlers) != 0

    def test_init_logger_exception_raise(self):
        with patch('mqtt_middlware_utils.logging_utils.init_logger',
                   side_effect=utils.LoggerInitException("Error initializing logger")):
            with pytest.raises(utils.LoggerInitException):
                utils.init_logger("LOG_LEVEL", "LOGGER_NAME",
                                  "LOG_DESTINATION")

    def test_log_info(self):
        log_level = "DEBUG"
        msg = "JUST A TEST FOR INFO LOGGING"
        logger_name = "test_logger"

        # delete log file if it exists
        self.clear_log_file()

        logger = utils.init_logger(
            log_level, logger_name, self.test_log_file)
        utils.log_info(logger_name, msg)
        written_line = self.read_file_line()

        assert "INFO" in written_line
        assert msg in written_line

    def test_log_info_exception_raise(self):
        with patch('mqtt_middlware_utils.logging_utils.log_info',
                   side_effect=utils.LoggerException("Error with logging info")):
            with pytest.raises(utils.LoggerException):
                log_level = "INFO"
                msg = "JUST A TEST MESSAGE"
                logger_name = "test_logger"

                self.clear_log_file()
                logger = utils.init_logger(
                    log_level, logger_name, self.test_log_file)
                utils.log_info(logger_name, msg)

    def test_log_debug(self):
        log_level = "DEBUG"
        msg = "JUST A TEST FOR DEBUG LOGGING"
        logger_name = "test_logger"

        # delete log file if it exists
        self.clear_log_file()

        logger = utils.init_logger(
            log_level, logger_name, self.test_log_file)
        utils.log_debug(logger_name, msg)
        written_line = self.read_file_line()

        assert "DEBUG" in written_line
        assert msg in written_line

    def test_log_debug_exception_raise(self):
        with patch('mqtt_middlware_utils.logging_utils.log_debug',
                   side_effect=utils.LoggerException("Error with logging debug")):
            with pytest.raises(utils.LoggerException):
                log_level = "DEBUG"
                msg = "JUST A TEST FOR DEBUG LOGGING"
                logger_name = "test_logger"

                self.clear_log_file()
                logger = utils.init_logger(
                    log_level, logger_name, self.test_log_file)
                utils.log_debug(logger_name, msg)

    def test_log_error(self):
        log_level = "DEBUG"
        msg = "JUST A TEST FOR ERROR LOGGING"
        logger_name = "test_logger"

        # delete log file if it exists
        self.clear_log_file()

        logger = utils.init_logger(
            log_level, logger_name, self.test_log_file)
        utils.log_error(logger_name, msg)
        written_line = self.read_file_line()

        assert "ERROR" in written_line
        assert msg in written_line

    def test_log_error_exception_raise(self):
        with patch('mqtt_middlware_utils.logging_utils.log_error',
                   side_effect=utils.LoggerException("Error with logging error")):
            with pytest.raises(utils.LoggerException):
                log_level = "DEBUG"
                msg = "JUST A TEST FOR ERROR LOGGING"
                logger_name = "test_logger"

                self.clear_log_file()
                logger = utils.init_logger(
                    log_level, logger_name, self.test_log_file)
                utils.log_error(logger_name, msg)

    def test_log_warning(self):
        log_level = "DEBUG"
        msg = "JUST A TEST FOR WARNING LOGGING"
        logger_name = "test_logger"

        # delete log file if it exists
        self.clear_log_file()

        logger = utils.init_logger(
            log_level, logger_name, self.test_log_file)
        utils.log_warning(logger_name, msg)
        written_line = self.read_file_line()

        assert "WARNING" in written_line
        assert msg in written_line

    def test_log_warning_exception_raise(self):
        with patch('mqtt_middlware_utils.logging_utils.log_warning',
                   side_effect=utils.LoggerException("Error with logging warning")):
            with pytest.raises(utils.LoggerException):
                log_level = "DEBUG"
                msg = "JUST A TEST FOR WARNING LOGGING"
                logger_name = "test_logger"

                self.clear_log_file()
                logger = utils.init_logger(
                    log_level, logger_name, self.test_log_file)
                utils.log_warning(logger_name, msg)




class MiddlewareConfigurationException(Exception):
    '''
    Raised when the provided configuration in not valid
    '''
    pass


class UuidGenerationError(Exception):
    '''
    Raised when an error occurs with the creation of an UUID
    '''
    pass


class SshRemoteError(Exception):
    '''
    Raised when an error occurs with a remote ssh connection session
    This can occurr while establishing a new connection or with the exection of a remote command
    '''
    pass


class SshCommandExecutionError(Exception):
    '''
    Raised when the execution of a remote ssh command fails
    '''
    pass


class DatabaseError(Exception):
    '''
    Generic database exception raised when an error occurs while interacting with the database
    '''
    pass


class DatabaseConnectionError(Exception):
    '''
    Raised when an error occurs while connecting to the database server
    '''
    pass


class DatabasePersistenceError(Exception):
    '''
    Raised when an error occurs while persisting an object to the database
    '''
    pass


class DatabaseUpdateError(Exception):
    '''
    Raised when an error occurs while updating an object
    '''
    pass


class DatabaseDeleteError(Exception):
    '''
    Raised when an error occurs while deleting an object
    '''
    pass


class DatabaseCreateIndexError(Exception):
    '''
    Raised when an error occurs when creating an index on the database
    '''
    pass


class DatabaseObjectNotFoundError(Exception):
    '''
    Raised when the requested object is not found in the database
    '''
    pass


class DatabaseObjectRetrieveError(Exception):
    '''
    Raised when an error occurs when retrieving an object from the database
    '''
    pass


class DatabaseIndexSearchError(Exception):
    '''
    Raised when an error occurs when searching the requested indexes
    '''
    pass


class LoggerInitException(Exception):
    '''
    Raised when an error occurs when initializing the logger
    '''
    pass


class LoggerException(Exception):
    '''
    Raised when a generic error occurs when logging.
    '''
    pass



import uuid
import configparser
import json

from fabric import Connection
from mqtt_middlware_utils.exceptions.mqtt_middleware_exceptions import MiddlewareConfigurationException, UuidGenerationError, SshRemoteError, SshCommandExecutionError


def generate_uuid():
    try:
        # generate uuid using random uuid generator
        new_instance_id = uuid.uuid4()
        return str(new_instance_id)
    except Exception as e:
        raise UuidGenerationError(
            "The following error occurred while generating a new UUID\n"+str(e))


def execute_ssh_command(conn, command) -> str:
    try:
        result = conn.run(command)
        if (result.ok):
            return result.stdout.strip()
        else:
            raise SshCommandExecutionError(
                "The following error occurred with the remote execution of the command ["+command+"]" + result.sterr.strip())
    except SshCommandExecutionError:
        raise
    except Exception as e:
        raise SshCommandExecutionError(
            "The following error occurred with the remote execution of the command ["+command+"]\n"+str(e))


def init_remote_ssh_connection(host, user, connection_kwargs):
    try:
        # TODO USE A SSH KEY INSTEAD! THIS IS NOT SAFE FOR PRODUCTION
        return Connection(host, user, connect_kwargs=connection_kwargs)
    except Exception as e:
        raise SshRemoteError(
            "The following error occurred while establishing a remote ssh connection with the host ["+host+"]\n"+str(e))


class middlware_instance:

    def __init__(self, instance_id, customer_id, status="offline") -> None:
        self.instance_id = instance_id
        self.customer_id = customer_id
        self.status = status


class mqtt_middleware_configuration:

    def __init__(self, instance_id="", sub_host="", sub_port=0, sub_topic="", source_client_id="", source_auth={}, pub_host="",
                 pub_port=0, pub_topic="", devices_auth_list={}, logger_name="", logger_level="", logger_dest="", *, dictionary={}) -> None:

        try:
            if not dictionary:
                self.mid_instance_id = instance_id
                self.sub_host = sub_host
                self.sub_port = sub_port
                self.sub_topic = sub_topic
                self.source_client_id = source_client_id
                self.source_auth = source_auth
                self.pub_host = pub_host
                self.pub_port = pub_port
                self.pub_topic = pub_topic
                self.devices_auth_list = devices_auth_list
                self.logger_name = logger_name
                self.logger_level = logger_level
                self.logger_dest = logger_dest
            else:
                self.__from_dictionary(dictionary)

            self.__checkValidConfiguration()
        except Exception as e:
            raise MiddlewareConfigurationException(
                "The following error occurred while creating mqtt_middleware_configuration: "
                + "\n"+str(e))

    def __from_dictionary(self, dictionary):
        "Create class from Python dictionary"
        self.mid_instance_id = dictionary["mid_instance_id"]
        self.sub_host = dictionary["sub_host"]
        self.sub_port = dictionary["sub_port"]
        self.sub_topic = dictionary["sub_topic"]
        self.source_client_id = dictionary["source_client_id"]
        self.source_auth = dictionary["source_auth"]
        self.pub_host = dictionary["pub_host"]
        self.pub_port = dictionary["pub_port"]
        self.pub_topic = dictionary["pub_topic"]
        self.devices_auth_list = dictionary["devices_auth_list"]
        self.logger_name = dictionary["logger_name"]
        self.logger_level = dictionary["logger_level"]
        self.logger_dest = dictionary["logger_dest"]

    def __checkValidConfiguration(self):
        def RaiseConfigurationException(config_line):
            raise MiddlewareConfigurationException("Invalid configuration provided" +
                                                   str(config_line)+" not set. Please check configuration")

        if not self.mid_instance_id:
            RaiseConfigurationException("Middleware Instance ID")
        if not self.sub_host:
            RaiseConfigurationException("Subscriber Host")
        if not self.sub_port:
            RaiseConfigurationException("Subscriber Port")
        if not self.sub_topic:
            RaiseConfigurationException("Subscriber Topic")
        if not self.source_client_id:
            RaiseConfigurationException("Subscriber Client_ID")
        if not self.source_auth:
            RaiseConfigurationException("Subscriber Auth Params")
        if not self.pub_host:
            RaiseConfigurationException("Publisher Host")
        if not self.pub_port:
            RaiseConfigurationException("Publisher Port")
        if not self.pub_topic:
            RaiseConfigurationException("Publisher Topic")
        if not self.devices_auth_list:
            RaiseConfigurationException("Publisher Auth Params")

    def debugPrintConfiguration(self):
        print("Issued Configuration")
        print("Logger level: " + self.logger_level)
        print("Logger name: "+self.logger_name)
        print("Logger destination: "+self.logger_dest)
        print("Subscriber Host: " + self.sub_host)
        print("Subscriber Port: " + str(self.sub_port))
        print("Subscriber Topic: " + self.sub_topic)
        print("Subscriber Client_ID: " + self.source_client_id)
        print("Subscriber Auth Params: " + str(self.source_auth))
        print("Publisher Host: "+self.pub_host)
        print("Publisher Port: "+str(self.pub_port))
        print("Publisher Topic: "+self.pub_topic)
        print("Publisher Auth Params: "+str(self.devices_auth_list))


def read_configuration(instance_id, config_file):
    conf_parser = configparser.ConfigParser()
    if not config_file:
        raise MiddlewareConfigurationException("Configuration file not issued")

    try:
        # read configuration file
        conf_parser.read(config_file)

        # get params from config file
        server_sub_config = conf_parser['SERVER.SUB']
        sub_host = server_sub_config["hostname"].strip()
        sub_port = server_sub_config.getint("port")
        sub_topic = server_sub_config["topic"].strip()
        source_client_id = server_sub_config["clientid"].strip()
        source_auth = json.loads(server_sub_config["auth"])

        server_pub_config = conf_parser["SERVER.PUB"]
        pub_host = server_pub_config["hostname"].strip()
        pub_port = server_pub_config.getint("port")
        pub_topic = server_pub_config["topic"].strip()
        devices_auth_list = json.loads(server_pub_config["auth"])

        server_sys_config = conf_parser["SYSTEM"]
        logger_name = server_sys_config["logname"]
        logger_level = server_sys_config["loglevel"]
        logger_dest = server_sys_config["logdestination"]

        return mqtt_middleware_configuration(instance_id, sub_host, sub_port, sub_topic, source_client_id,
                                             source_auth, pub_host, pub_port, pub_topic, devices_auth_list,
                                             logger_name, logger_level, logger_dest)
    except MiddlewareConfigurationException as e:
        raise
    except Exception as e:
        raise MiddlewareConfigurationException(
            "The following error occurred while reading configuration from file. Please check the provided configuration keys\n"+str(e))



import pytest

from mqtt_middlware_utils import utils
from unittest.mock import patch, Mock


class TestUtilsModule:

    def test_generate_uuid(self):
        uuid = utils.generate_uuid()
        assert (len(uuid) == 36)

    def test_generate_uuid_string_type(self):
        uuid = utils.generate_uuid()
        assert (isinstance(uuid, str))

    @patch('mqtt_middlware_utils.utils.uuid.uuid4')
    def test_generate_uuid_exception_raise(self, mocked_uuid4):
        with pytest.raises(utils.UuidGenerationError) as e:
            mocked_uuid4.side_effect = ValueError('A generic error occurred')
            utils.generate_uuid()
        error_string = str(e.value)
        assert "The following error occurred while generating a new UUID" in error_string
        assert "A generic error occurred" in error_string
        mocked_uuid4.assert_called_once()

    @patch('mqtt_middlware_utils.utils.init_remote_ssh_connection')
    def test_execute_ssh_command(self, mocked_init_remote_ssh_connection):
        cmd = "echo test"
        cmd_stdout = "test"

        mocked_result = Mock(ok=True, stdout=cmd_stdout)
        mocked_connection = Mock()
        mocked_connection.run.return_value = mocked_result

        mocked_init_remote_ssh_connection.return_value = mocked_connection

        # first get connection
        ssh_connection = utils.init_remote_ssh_connection(
            "test_host", "user", connection_kwargs={"password": "password"})

        cmd_result = utils.execute_ssh_command(ssh_connection, cmd)

        assert cmd_result == cmd_stdout
        mocked_connection.run.assert_called_once_with(cmd)

    @patch('mqtt_middlware_utils.utils.init_remote_ssh_connection')
    def test_execute_ssh_command_result_not_ok(self, mocked_init_remote_ssh_connection):
        with pytest.raises(utils.SshCommandExecutionError) as e:
            cmd = "echo test"
            cmd_stdout = "test"

            mocked_result = Mock(ok=False, stdout=cmd_stdout)
            mocked_connection = Mock()
            mocked_connection.run.return_value = mocked_result

            mocked_init_remote_ssh_connection.return_value = mocked_connection

            # first get connection
            ssh_connection = utils.init_remote_ssh_connection(
                "test_host", "user", connection_kwargs={"password": "password"})

            utils.execute_ssh_command(ssh_connection, cmd)

        error_string = str(e.value)
        assert "The following error occurred with the remote execution of the command" in error_string
        assert cmd in error_string
        mocked_connection.run.assert_called_once_with(cmd)

    @patch('mqtt_middlware_utils.utils.init_remote_ssh_connection')
    def test_execute_ssh_command_exception_raise(self, mocked_init_remote_ssh_connection):
        with pytest.raises(utils.SshCommandExecutionError) as e:
            cmd = "echo test"

            mocked_connection = Mock()
            mocked_connection.run.side_effect = ValueError(
                'A generic error occurred')

            mocked_init_remote_ssh_connection.return_value = mocked_connection

            # first get connection
            ssh_connection = utils.init_remote_ssh_connection(
                "test_host", "user", connection_kwargs={"password": "password"})

            utils.execute_ssh_command(ssh_connection, cmd)

        error_string = str(e.value)
        assert "The following error occurred with the remote execution of the command" in error_string
        assert "A generic error occurred" in error_string
        assert cmd in error_string
        mocked_connection.run.assert_called_once_with(cmd)

    @patch('mqtt_middlware_utils.utils.Connection', autospec=True)
    def test_init_remote_ssh_connection(self, mocked_connection):
        connect_return = Mock()
        connect_return.is_connected.return_value = True
        mocked_connection.return_value = connect_return

        host = "test_host"
        user = "test_user"
        connection_kwargs = {"password": "test_password"}

        conn = utils.init_remote_ssh_connection(
            host, user, connection_kwargs=connection_kwargs)

        assert conn.is_connected() == True
        mocked_connection.assert_called_once_with(
            host, user, connect_kwargs=connection_kwargs)

    @patch('mqtt_middlware_utils.utils.Connection', autospec=True)
    def test_init_remote_ssh_connection_not_connected(self, mocked_connection):
        connect_return = Mock()
        connect_return.is_connected.return_value = False
        mocked_connection.return_value = connect_return

        host = "test_host"
        user = "test_user"
        connection_kwargs = {"password": "test_password"}

        conn = utils.init_remote_ssh_connection(
            host, user, connection_kwargs=connection_kwargs)

        assert conn.is_connected() == False
        mocked_connection.assert_called_once_with(
            host, user, connect_kwargs=connection_kwargs)

    @patch('mqtt_middlware_utils.utils.Connection', autospec=True, side_effect=ValueError("Error establishing connection"))
    def test_init_remote_ssh_connection_exception(self, mocked_connection):
        with pytest.raises(utils.SshRemoteError) as e:

            host = "test_host"
            user = "test_user"
            connection_kwargs = {"password": "test_password"}

            utils.init_remote_ssh_connection(
                host, user, connection_kwargs=connection_kwargs)

        error_string = str(e.value)
        assert "The following error occurred while establishing a remote ssh connection with the host" in error_string
        assert "Error establishing connection" in error_string
        assert host in error_string
        mocked_connection.assert_called_once_with(
            host, user, connect_kwargs=connection_kwargs)

    def test_middlware_instance_init(self):
        instance_id = "test-test1"
        customer_id = "0001"
        status = "offline"
        mid_instance = utils.middlware_instance(
            instance_id, customer_id, status)

        assert mid_instance.instance_id == "test-test1"
        assert mid_instance.customer_id == "0001"
        assert mid_instance.status == "offline"

    def test_mqtt_middleware_configuration_init(self):
        instance_id = "test-test1"
        sub_host = "host1"
        sub_port = 8080
        sub_topic = "test_topic"
        source_client_id = "TEST-CLIENT"
        source_auth = {"username": "user", "password": "user"}
        pub_host = "host2"
        pub_port = 8080
        pub_topic = "test_topic2"
        devices_auth_list = {
            "0": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            },
            "1": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            }}
        logger_name = "logger"
        logger_level = "DEBUG"
        logger_dest = "test/path/to/logs"

        mid_config = utils.mqtt_middleware_configuration(instance_id, sub_host, sub_port, sub_topic, source_client_id, source_auth, pub_host,
                                                         pub_port, pub_topic, devices_auth_list, logger_name, logger_level, logger_dest)

        assert mid_config.mid_instance_id == "test-test1"
        assert mid_config.sub_host == "host1"
        assert mid_config.sub_port == 8080
        assert mid_config.sub_topic == "test_topic"
        assert mid_config.source_client_id == "TEST-CLIENT"
        assert mid_config.source_auth["username"] == "user"
        assert mid_config.source_auth["password"] == "user"
        assert mid_config.pub_host == "host2"
        assert mid_config.pub_port == 8080
        assert mid_config.pub_topic == "test_topic2"
        assert mid_config.devices_auth_list["0"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["0"]["username"] == "user"
        assert mid_config.devices_auth_list["0"]["password"] == "user"
        assert mid_config.devices_auth_list["1"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["1"]["username"] == "user"
        assert mid_config.devices_auth_list["1"]["password"] == "user"
        assert mid_config.logger_name == "logger"
        assert mid_config.logger_level == "DEBUG"
        assert mid_config.logger_dest == "test/path/to/logs"

    def test_mqtt_middleware_configuration_dictionary_init(self):
        config_dict = {}
        config_dict["mid_instance_id"] = "test-test1"
        config_dict["sub_host"] = "host1"
        config_dict["sub_port"] = 8080
        config_dict["sub_topic"] = "test_topic"
        config_dict["source_client_id"] = "TEST-CLIENT"
        config_dict["source_auth"] = {"username": "user", "password": "user"}
        config_dict["pub_host"] = "host2"
        config_dict["pub_port"] = 8080
        config_dict["pub_topic"] = "test_topic2"
        config_dict["devices_auth_list"] = {
            "0": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            },
            "1": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            }}
        config_dict["logger_name"] = "logger"
        config_dict["logger_level"] = "DEBUG"
        config_dict["logger_dest"] = "test/path/to/logs"

        mid_config = utils.mqtt_middleware_configuration(
            dictionary=config_dict)

        assert mid_config.mid_instance_id == "test-test1"
        assert mid_config.sub_host == "host1"
        assert mid_config.sub_port == 8080
        assert mid_config.sub_topic == "test_topic"
        assert mid_config.source_client_id == "TEST-CLIENT"
        assert mid_config.source_auth["username"] == "user"
        assert mid_config.source_auth["password"] == "user"
        assert mid_config.pub_host == "host2"
        assert mid_config.pub_port == 8080
        assert mid_config.pub_topic == "test_topic2"
        assert mid_config.devices_auth_list["0"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["0"]["username"] == "user"
        assert mid_config.devices_auth_list["0"]["password"] == "user"
        assert mid_config.devices_auth_list["1"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["1"]["username"] == "user"
        assert mid_config.devices_auth_list["1"]["password"] == "user"
        assert mid_config.logger_name == "logger"
        assert mid_config.logger_level == "DEBUG"
        assert mid_config.logger_dest == "test/path/to/logs"

    def test_mqtt_middleware_configuration_dictionary_init_double_constructor(self):
        config_dict = {}
        config_dict["mid_instance_id"] = "test-test1"
        config_dict["sub_host"] = "host1"
        config_dict["sub_port"] = 8080
        config_dict["sub_topic"] = "test_topic"
        config_dict["source_client_id"] = "TEST-CLIENT"
        config_dict["source_auth"] = {"username": "user", "password": "user"}
        config_dict["pub_host"] = "host2"
        config_dict["pub_port"] = 8080
        config_dict["pub_topic"] = "test_topic2"
        config_dict["devices_auth_list"] = {
            "0": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            },
            "1": {
                "clientid": "test_id",
                "username": "user",
                "password": "user"
            }}
        config_dict["logger_name"] = "logger"
        config_dict["logger_level"] = "DEBUG"
        config_dict["logger_dest"] = "test/path/to/logs"

        # these should be different from the final config
        instance_id = "test-test2"
        sub_host = "host2"
        sub_port = 8081
        sub_topic = "test_topic_2"
        source_client_id = "TEST-CLIENT-2"
        source_auth = {"username": "user2", "password": "user22"}
        pub_host = "host3"
        pub_port = 8082
        pub_topic = "test_topic3"
        devices_auth_list = {
            "0": {
                "clientid": "test_id_1",
                "username": "user",
                "password": "user"
            },
            "1": {
                "clientid": "test_id_2",
                "username": "user",
                "password": "user"
            }}
        logger_name = "logger_2"
        logger_level = "NOT_DEBUG"
        logger_dest = "test/path/to/logs_different"

        mid_config = utils.mqtt_middleware_configuration(instance_id, sub_host, sub_port, sub_topic, source_client_id, source_auth, pub_host,
                                                         pub_port, pub_topic, devices_auth_list, logger_name, logger_level, logger_dest, dictionary=config_dict)

        assert mid_config.mid_instance_id == "test-test1"
        assert mid_config.sub_host == "host1"
        assert mid_config.sub_port == 8080
        assert mid_config.sub_topic == "test_topic"
        assert mid_config.source_client_id == "TEST-CLIENT"
        assert mid_config.source_auth["username"] == "user"
        assert mid_config.source_auth["password"] == "user"
        assert mid_config.pub_host == "host2"
        assert mid_config.pub_port == 8080
        assert mid_config.pub_topic == "test_topic2"
        assert mid_config.devices_auth_list["0"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["0"]["username"] == "user"
        assert mid_config.devices_auth_list["0"]["password"] == "user"
        assert mid_config.devices_auth_list["1"]["clientid"] == "test_id"
        assert mid_config.devices_auth_list["1"]["username"] == "user"
        assert mid_config.devices_auth_list["1"]["password"] == "user"
        assert mid_config.logger_name == "logger"
        assert mid_config.logger_level == "DEBUG"
        assert mid_config.logger_dest == "test/path/to/logs"

    def test_mqtt_middleware_configuration_bad_config(self):
        with pytest.raises(utils.MiddlewareConfigurationException):
            sub_host = "host1"
            sub_port = 8080

            utils.mqtt_middleware_configuration(
                sub_host=sub_host, sub_port=sub_port)

    def test_mqtt_middleware_configuration_dictionary_bad_configuration(self):
        with pytest.raises(utils.MiddlewareConfigurationException):
            config_dict = {}
            config_dict["mid_instance_id"] = "test-test1"
            config_dict["sub_host"] = "host1"
            config_dict["sub_port"] = 8080

            utils.mqtt_middleware_configuration(
                dictionary=config_dict)

    def test_read_configuration(self):
        config_file = "tests/files/config.ini"
        mid_instance_id = "instance-test"
        mid_config = utils.read_configuration(mid_instance_id, config_file)

        assert mid_config.mid_instance_id == "instance-test"
        assert mid_config.sub_host == "host"
        assert mid_config.sub_port == 8080
        assert mid_config.sub_topic == "test_topic"
        assert mid_config.source_client_id == "CLIENT-TEST"
        assert mid_config.source_auth["username"] == "user"
        assert mid_config.source_auth["password"] == "user"
        assert mid_config.pub_host == "host_2"
        assert mid_config.pub_port == 8080
        assert mid_config.pub_topic == "test_topic"
        assert mid_config.devices_auth_list["0"]["clientid"] == "client-1"
        assert mid_config.devices_auth_list["0"]["username"] == "user"
        assert mid_config.devices_auth_list["0"]["password"] == "user"
        assert mid_config.devices_auth_list["1"]["clientid"] == "client-2"
        assert mid_config.devices_auth_list["1"]["username"] == "user"
        assert mid_config.devices_auth_list["1"]["password"] == "user"
        assert mid_config.logger_name == "mqtt_integration"
        assert mid_config.logger_level == "DEBUG"
        assert mid_config.logger_dest == "/var/log/test/test.log"

    def test_read_configuration_exception_raise(self):
        with pytest.raises(utils.MiddlewareConfigurationException):
            config_file = "tests/files/wrong_config.ini"
            mid_instance_id = "instance-test"
            utils.read_configuration(mid_instance_id, config_file)



import pytest

from unittest import TestCase
from unittest.mock import patch, Mock

import mqtt_middlware_utils.db_utils as utils


def simple_function_for_decorator_testing():
    pass


class TestDbUtils(TestCase):

    @patch('mqtt_middlware_utils.db_utils.redis.Redis', autospec=True)
    def test_get_new_connection(self, mock_redis_class):
        # mock redis.Redis in get_new_connection() in order to return a fake connection
        self.assertIs(utils.redis.Redis, mock_redis_class)
        mock_redis_class.return_value = "Connected"

        host = "localhost"
        username = "user"
        password = "user"

        redis_conn = utils.get_new_connection(host, username, password)

        assert redis_conn == "Connected"

    @patch('mqtt_middlware_utils.db_utils.redis.Redis', autospec=True)
    def test_get_new_connection_exception(self, mock_redis_class):
        with pytest.raises(utils.DatabaseConnectionError) as e:

            self.assertIs(utils.redis.Redis, mock_redis_class)
            mock_redis_class.side_effect = ValueError(
                "Something went wrong during connection")

            host = "localhost"
            username = "user"
            password = "user"

            utils.get_new_connection(host, username, password)
        assert "The following error occurred trying to connect to database" in str(
            e)
        assert "Something went wrong during connection" in str(e)

    @patch('mqtt_middlware_utils.db_utils.redis.Redis', autospec=True)
    def test_get_new_connection_no_login(self, mock_redis_class):
        # mock redis.Redis in get_new_connection() in order to return a fake connection
        self.assertIs(utils.redis.Redis, mock_redis_class)
        mock_redis_class.return_value = "Connected"

        host = "localhost"

        redis_conn = utils.get_new_connection_no_login(
            host)

        assert redis_conn == "Connected"

    @patch('mqtt_middlware_utils.db_utils.redis.Redis', autospec=True)
    def test_get_new_connection_no_login_exception(self, mock_redis_class):
        with pytest.raises(utils.DatabaseConnectionError) as e:

            self.assertIs(utils.redis.Redis, mock_redis_class)
            mock_redis_class.side_effect = ValueError(
                "Something went wrong during connection")

            host = "localhost"

            utils.get_new_connection_no_login(host)
        assert "The following error occurred trying to connect to database" in str(
            e)
        assert "Something went wrong during connection" in str(e)

    # TODO implement this test
    def test_get_new_connection_ssl(self):
        pass

    # TODO implement this test
    def test_get_new_connection_ssl_exception(self):
        pass

    # TODO implement this test
    def test_get_new_connection_no_login_ssl(self):
        pass

    # TODO implement this test
    def test_get_new_connection_no_login_ssl_exception(self):
        pass

    @patch('mqtt_middlware_utils.db_utils.retrieve_object', autospec=True)
    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_persist_new_object(self, mocked_get_new_connection, mocked_retrieve_object):
        object_to_persist = {"name": "John", "age": 30}

        mocked_connection = Mock()
        mocked_connection.json().set.return_value = "OK"
        mocked_get_new_connection.return_value = mocked_connection

        mocked_retrieve_object.return_value = object_to_persist

        host = "localhost"
        username = "test"
        password = "test"

        object_name = "test_object"

        # get a new connection
        connection = utils.get_new_connection(
            host=host, username=username, password=password)

        persisted_object = utils.persist_new_object(
            connection, object_name, object_to_persist)

        assert persisted_object == object_to_persist
        mocked_connection.json().set.assert_called_once_with(
            object_name, '$', object_to_persist)
        mocked_retrieve_object.assert_called_once_with(
            connection, object_name)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_persist_new_object_result_not_ok(self, mocked_get_new_connection):
        object_name = "test_object"
        with pytest.raises(utils.DatabasePersistenceError) as e:

            mocked_connection = Mock()
            mocked_connection.json().set.return_value = "nil"
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            utils.persist_new_object(
                connection, object_name, {"name": "John", "age": 30})

        assert "Persist operation resulted NOT OK" in str(e)
        assert object_name in str(e)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_persist_new_object_json_set_exception(self, mocked_get_new_connection):
        object_name = "test_object"
        with pytest.raises(utils.DatabasePersistenceError) as e:
            object_to_persist = {"name": "John", "age": 30}

            mocked_connection = Mock()
            mocked_connection.json().set.side_effect = ValueError("A generic error occurred")
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="host", username="username", password="password")

            utils.persist_new_object(
                connection, object_name, object_to_persist)

        assert "The following error occurred trying to persist the object" in str(
            e)
        assert object_name in str(e)
        assert "A generic error occurred" in str(e)

    @patch('mqtt_middlware_utils.db_utils.persist_new_object', side_effect=utils.DatabasePersistenceError("A generic error occurred"), autospec=True)
    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_persist_new_object_generic_exception(self, mocked_get_new_connection, mocked_persist_new_object):
        with pytest.raises(utils.DatabasePersistenceError):
            mocked_connection = Mock()
            mocked_connection.json().set.return_value = "OK"
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            # keep the decorator for the mocked function
            mocked_persist_new_object(connection, "test_object", {
                                      "name": "John", "age": 30})

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_retrieve_object(self, mocked_get_new_connection):

        object_to_retrieve = {"name": "John", "age": 30}
        object_name = "test_object"

        mocked_connection = Mock()
        mocked_connection.json().get.return_value = object_to_retrieve
        mocked_get_new_connection.return_value = mocked_connection

        # get a new connection
        connection = utils.get_new_connection(
            host="localhost", username="test", password="test")

        retrieved_object = utils.retrieve_object(
            connection, object_name)

        assert retrieved_object == object_to_retrieve
        mocked_connection.json().get.assert_called_once_with(
            object_name)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_retrieve_object_object_path(self, mocked_get_new_connection):

        object_to_retrieve = {"name": "John", "age": 30}
        object_name = "test_object"
        object_path = '$'

        mocked_connection = Mock()
        mocked_connection.json().get.return_value = object_to_retrieve
        mocked_get_new_connection.return_value = mocked_connection

        # get a new connection
        connection = utils.get_new_connection(
            host="localhost", username="test", password="test")

        retrieved_object = utils.retrieve_object(
            connection, object_name, object_path)

        assert retrieved_object == object_to_retrieve

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_retrieve_object_generic_exception(self, mocked_get_new_connection):
        with pytest.raises(utils.DatabaseObjectRetrieveError) as e:
            mocked_connection = Mock()
            mocked_connection.json().get.side_effect = ValueError("A generic error occurred")
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            object_name = "test_object"

            utils.retrieve_object(connection, object_name)
        assert "The following error occurred trying to retrieve the object" in str(
            e)
        assert "A generic error occurred" in str(e)
        assert object_name in str(e)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_create_index(self, mocked_get_new_connection):
        mocked_index = Mock()
        mocked_index.create_index.return_value = "OK"

        mocked_connection = Mock()
        mocked_connection.ft.return_value = mocked_index
        mocked_get_new_connection.return_value = mocked_connection

        # get a new connection
        connection = utils.get_new_connection(
            host="localhost", username="test", password="test")

        index_name = "test_index"
        schema = "test_schema"
        definition = "test_definition"
        index_type = "test_index_type"

        created_idx = utils.create_index(connection, index_name,
                                         schema, definition, index_type)

        assert created_idx == mocked_index
        mocked_connection.ft.assert_called_once_with(index_name)
        mocked_index.create_index.assert_called_once_with(
            schema, definition, index_type)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_create_index_not_ok_exception(self, mocked_get_new_connection):
        with pytest.raises(utils.DatabaseCreateIndexError) as e:
            mocked_index = Mock()
            mocked_index.create_index.result = "ERROR"

            mocked_connection = Mock()
            mocked_connection.ft.return_value = mocked_index
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            index_name = "test_index"
            schema = "test_schema"
            definition = "test_definition"
            index_type = "test_index_type"

            utils.create_index(connection, index_name,
                               schema, definition, index_type)
        error_string = str(e.value)
        assert "Create index operation resulted NOT OK for index" in error_string
        assert index_name in error_string
        assert schema in error_string
        assert definition in error_string
        assert index_type in error_string

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_create_index_generic_exception(self, mocked_get_new_connection):
        with pytest.raises(utils.DatabaseCreateIndexError) as e:
            mocked_index = Mock()
            mocked_index.create_index.side_effect = ValueError(
                "A generic error occurred")

            mocked_connection = Mock()
            mocked_connection.ft.return_value = mocked_index
            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            index_name = "test_index"
            schema = "test_schema"
            definition = "test_definition"
            index_type = "test_index_type"

            utils.create_index(connection, index_name,
                               schema, definition, index_type)
        error_string = str(e.value)
        assert "The following error occurred trying to create the index" in error_string
        assert "A generic error occurred" in error_string
        assert index_name in error_string
        assert schema in error_string
        assert definition in error_string
        assert index_type in error_string

    @patch('mqtt_middlware_utils.db_utils.create_index', autospec=True)
    def test_search_from_index(self, mocked_create_index):
        expected_search_result = "test_search_result"

        mocked_index = Mock()
        mocked_index.search.return_value = expected_search_result
        mocked_create_index.return_value = mocked_index

        mocked_connection = Mock()

        index_name = "test_index"
        schema = "test_schema"
        definition = "test_definition"
        index_type = "test_index_type"

        # create the index first
        idx = utils.create_index(mocked_connection, index_name,
                                 schema, definition, index_type)

        query_string = "test_query"

        search_result = utils.search_from_index(idx, query_string)

        assert search_result == expected_search_result
        mocked_index.search.assert_called_once_with(query_string)

    @patch('mqtt_middlware_utils.db_utils.create_index', autospec=True)
    def test_search_from_index_exception(self, mocked_create_index):
        with pytest.raises(utils.DatabaseIndexSearchError) as e:
            index_name = "test_index"
            schema = "test_schema"
            definition = "test_definition"
            index_type = "test_index_type"

            mocked_index = Mock()
            mocked_index.name = index_name
            mocked_index.search.side_effect = ValueError(
                "A generic error occurred")
            mocked_create_index.return_value = mocked_index

            mocked_connection = Mock()

            # create the index first
            idx = utils.create_index(mocked_connection, index_name,
                                     schema, definition, index_type)

            query_string = "test_query"

            utils.search_from_index(idx, query_string)
        error_string = str(e.value)
        assert "The following error occurred trying to search the index" in error_string
        assert "A generic error occurred" in error_string
        assert index_name in error_string
        assert query_string in error_string
        mocked_index.search.assert_called_once_with(query_string)

    @patch('mqtt_middlware_utils.db_utils.persist_new_object', autospec=True)
    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_update_object(self, mocked_get_new_connection, mocked_persist_new_object):
        mocked_connection = Mock()
        mocked_get_new_connection.return_value = mocked_connection

        object_to_persist = {"name": "John", "age": 30}
        object_name = "test_object"

        mocked_persist_new_object.return_value = object_to_persist

        # get a new connection
        connection = utils.get_new_connection(
            host="localhost", username="test", password="test")

        updated_object = utils.update_object(
            connection, object_name, object_to_persist)

        assert updated_object == object_to_persist
        mocked_persist_new_object.assert_called_once_with(
            connection, object_name, object_to_persist)

    @patch('mqtt_middlware_utils.db_utils.persist_new_object', autospec=True)
    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_update_object_exception(self, mocked_get_new_connection, mocked_persist_new_object):
        with pytest.raises(utils.DatabaseUpdateError) as e:
            mocked_connection = Mock()
            mocked_get_new_connection.return_value = mocked_connection

            object_to_update = {"name": "John", "age": 30}
            object_name = "test_object"

            mocked_persist_new_object.side_effect = ValueError(
                "A generic error occurred")

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            utils.update_object(
                connection, object_name, object_to_update)
        error_string = str(e.value)
        assert "The following error occurred trying to update the object" in error_string
        assert object_name in error_string
        assert str(object_to_update) in error_string
        mocked_persist_new_object.assert_called_once_with(
            connection, object_name, object_to_update)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_delete_object(self, mocked_get_new_connection):
        mocked_connection = Mock()
        mocked_connection.json().delete.return_value = 1

        mocked_get_new_connection.return_value = mocked_connection

        # get a new connection
        connection = utils.get_new_connection(
            host="localhost", username="test", password="test")

        object_name = "test_object"

        deleted_objects = utils.delete_object(connection, object_name)

        assert deleted_objects == 1
        mocked_connection.json().delete.assert_called_once_with(object_name)

    @patch('mqtt_middlware_utils.db_utils.get_new_connection', autospec=True)
    def test_delete_object_exception(self, mocked_get_new_connection):
        with pytest.raises(utils.DatabaseDeleteError) as e:
            mocked_connection = Mock()
            mocked_connection.json().delete.side_effect = ValueError("A generic error occurred")

            mocked_get_new_connection.return_value = mocked_connection

            # get a new connection
            connection = utils.get_new_connection(
                host="localhost", username="test", password="test")

            object_name = "test_object"

            utils.delete_object(connection, object_name)

        error_string = str(e.value)
        assert "The following error occurred trying to delete the object" in error_string
        assert object_name in error_string
        mocked_connection.json().delete.assert_called_once_with(object_name)

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_generic_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseError) as e:
            mocked_test_function.side_effect = ValueError(
                "A generic error occurred")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "The following error occurred while operating on the database" in error_string
        assert "A generic error occurred" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_object_not_found_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseObjectNotFoundError) as e:
            mocked_test_function.side_effect = utils.DatabaseObjectNotFoundError(
                "Can't find the object")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "Can't find the object" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_connection_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseConnectionError) as e:
            mocked_test_function.side_effect = utils.DatabaseConnectionError(
                "An error occurred while connecting to the database")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while connecting to the database" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_persistence_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabasePersistenceError) as e:
            mocked_test_function.side_effect = utils.DatabasePersistenceError(
                "An error occurred while persisting an object")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while persisting an object" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_update_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseUpdateError) as e:
            mocked_test_function.side_effect = utils.DatabaseUpdateError(
                "An error occurred while updating an object")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while updating an object" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_delete_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseDeleteError) as e:
            mocked_test_function.side_effect = utils.DatabaseDeleteError(
                "An error occurred while deleting an object")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while deleting an object" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_create_index_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseCreateIndexError) as e:
            mocked_test_function.side_effect = utils.DatabaseCreateIndexError(
                "An error occurred while creating an index")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while creating an index" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_object_retrieve_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseObjectRetrieveError) as e:
            mocked_test_function.side_effect = utils.DatabaseObjectRetrieveError(
                "An error occurred while retrieving an object")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while retrieving an object" in error_string
        mocked_test_function.assert_called_once()

    @patch('tests.test_db_utils.simple_function_for_decorator_testing', autospec=True)
    def test_function_wrapper_index_search_exception(self, mocked_test_function):
        with pytest.raises(utils.DatabaseIndexSearchError) as e:
            mocked_test_function.side_effect = utils.DatabaseIndexSearchError(
                "An error occurred while searching an index")
            decorated_function = utils.db_exceptions_decorator(
                mocked_test_function)
            decorated_function()
        error_string = str(e.value)
        assert "An error occurred while searching an index" in error_string
        mocked_test_function.assert_called_once()



import redis

from redis.commands.json.path import Path

from mqtt_middlware_utils.exceptions.mqtt_middleware_exceptions import DatabaseError, DatabaseObjectNotFoundError, DatabaseConnectionError, DatabasePersistenceError, DatabaseCreateIndexError, DatabaseUpdateError, DatabaseDeleteError, DatabaseObjectRetrieveError, DatabaseIndexSearchError


def db_exceptions_decorator(func):
    '''
    This decorator function avoids code duplication between database functions for error handling.

    @param func: the function to be decorated   

    @returns: the return value of the decorated function

    @raise DatabaseObjectNotFoundError if the desired database object is not found.
    @raise DatabaseConnectionError if an error occurrs during database connection.
    @raise DatabasePersistenceError if an error occurrs during database persistence.
    @raise DatabaseUpdateError if an error occurrs during database update.
    @raise DatabaseDeleteError if an error occurrs during database deletion.
    @raise DatabaseCreateIndexError if an error occurrs during database creation of an index.
    @raise DatabaseObjectRetrieveError if an error occurrs during database retrieval of an object.
    @raise DatabaseIndexSearchError if an error occurrs during database search of an index.s
    @raise DatabaseError if a generic error occurrs during database operations.

    '''
    def function_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DatabaseObjectNotFoundError:
            raise
        except DatabaseConnectionError:
            raise
        except DatabasePersistenceError:
            raise
        except DatabaseUpdateError:
            raise
        except DatabaseDeleteError:
            raise
        except DatabaseCreateIndexError:
            raise
        except DatabaseObjectRetrieveError:
            raise
        except DatabaseIndexSearchError:
            raise
        except Exception as e:
            raise DatabaseError(
                "The following error occurred while operating on the database: \n"+str(e))
    return function_wrapper


@db_exceptions_decorator
def get_new_connection(host, username, password, port=6379, ssl=False, ssl_certfile=None,
                       ssl_keyfile=None, ssl_ca_certs=None, decode_responses=False):
    '''
    Get a new connection to a redis server using the provided credentials.

    @param host: the hostname of the redis server
    @param username: the username to connect to the redis server
    @param password: the password to connect to the redis server
    @param port: the port to connect to the redis server
    @param ssl: whether or not to use ssl
    @param ssl_certfile: the path to the ssl certificate file
    @param ssl_keyfile: the path to the ssl key file
    @param ssl_ca_certs: the path to the ssl certificate authority file
    @param decode_responses: whether or not to decode responses from the redis server

    @return: a new redis connection

    @raise DatabaseConnectionError if an error occurrs during database connection.  
    '''
    try:
        conn = None
        if (ssl):
            conn = redis.Redis(host=host, port=port, username=username, password=password,
                               ssl=ssl, ssl_certfile=ssl_certfile, ssl_keyfile=ssl_keyfile,
                               ssl_ca_certs=ssl_ca_certs, decode_responses=decode_responses)
        else:
            conn = redis.Redis(host=host, port=port, username=username, password=password,
                               decode_responses=decode_responses)
        return conn
    except Exception as e:
        raise DatabaseConnectionError(
            "The following error occurred trying to connect to database with params [ (host: " +
            str(host)+") (port: " + str(port) + ") (username: " + str(username) + ") (password: "+str(password) +
            ") (ssl: "+str(ssl)+") (ssl_certfile: "+str(ssl_certfile)+") (ssl_keyfile: "+str(ssl_keyfile) +
            ") (ssl_ca_certs: "+str(ssl_ca_certs)+") (decode_responses: "+str(decode_responses)+") ]: \n"+str(e))


@db_exceptions_decorator
def get_new_connection_no_login(host, port=6379, ssl=False, ssl_certfile=None,
                                ssl_keyfile=None, ssl_ca_certs=None, decode_responses=False):
    '''
    Get a new connection to a redis server using the default credentials.

    @param host: the hostname of the redis server
    @param port: the port to connect to the redis server
    @param ssl: whether or not to use ssl
    @param ssl_certfile: the path to the ssl certificate file  
    @param ssl_keyfile: the path to the ssl key file
    @param ssl_ca_certs: the path to the ssl certificate authority file
    @param decode_responses: whether or not to decode responses from the redis server 

    @return: a new redis connection 

    @raise DatabaseConnectionError if an error occurrs during database connection.
    '''
    try:
        conn = None
        if (ssl):
            conn = redis.Redis(host=host, port=port, ssl=ssl, ssl_certfile=ssl_certfile,
                               ssl_keyfile=ssl_keyfile, ssl_ca_certs=ssl_ca_certs,
                               decode_responses=decode_responses)
        else:
            conn = redis.Redis(host=host, port=port,
                               decode_responses=decode_responses)
        return conn
    except Exception as e:
        raise DatabaseConnectionError(
            "The following error occurred trying to connect to database with params [ (host: " +
            str(host)+") (port: " + str(port) + ") (ssl: "+str(ssl)+") (ssl_certfile: "+str(ssl_certfile) +
            ") (ssl_keyfile: "+str(ssl_keyfile) + ") (ssl_ca_certs: " +
            str(ssl_ca_certs)+") (decode_responses: " + str(decode_responses)+") ]: \n"+str(e))


@db_exceptions_decorator
def persist_new_object(connection, object_name, new_object, source='$'):
    '''
    Persist a new object in the database using the provided connection.

    @param connection: the redis connection to use
    @param object_name: the name of the object to persist
    @param new_object: the object to persist
    @param source: the source of the object to persist

    @return: the persisted object

    @raise DatabasePersistenceError if an error occurrs during the persist operation.
    '''

    try:
        result = connection.json().set(object_name, source, new_object)
        # get the newly persisted object from the database
        return retrieve_object(connection, object_name)
    except Exception as e:
        raise DatabasePersistenceError(
            "The following error occurred trying to persist the object [(object_name: " +
            str(object_name) + ") (source: "+str(source)+")]: \n"+str(e))


@db_exceptions_decorator
def retrieve_object(connection, object_name, object_path=None):
    '''
    Retrieves an object from the database using the provided connection.

    @param connection: the redis connection to use
    @param object_name: the name of the object to retrieve
    @param object_path: the path to the object to retrieve

    @return: the retrieved object

    @raise DatabaseObjectRetrieveError if an error occurrs during the retrieve operation.
    '''
    try:
        result = None
        if (object_path is not None):
            result = connection.json().get(object_name, Path('.'+object_path))
        else:
            result = connection.json().get(object_name)
        return result
    except Exception as e:
        raise DatabaseObjectRetrieveError(
            "The following error occurred trying to retrieve the object [(object_name: " + str(object_name) + ")]: \n"+str(e))


@db_exceptions_decorator
def create_index(conn, index_name, schema, definition, index_type):
    '''
    Creates an index using the provided connection.

    @param connection: the redis connection to use
    @param index_name: the name of the index to create
    @param schema: the schema of the index to create
    @param definition: the definition of the index to create
    @param index_type: the type of the index to create

    @return: the created index

    @raise DatabaseCreateIndexError if an error occurrs during the create index operation.
    '''
    try:
        idx = conn.ft(index_name)
        result = idx.create_index(schema, definition, index_type)
        if (result != "OK"):
            raise DatabaseCreateIndexError(
                "Create index operation resulted NOT OK for index [(index_name: " +
                str(index_name) + ") (schema: "+str(schema)+") (definition: "+str(definition) +
                ") (index_type: "+str(index_type)+")]: \n"+str(result))
        return idx
    except DatabaseCreateIndexError:
        raise
    except Exception as e:
        raise DatabaseCreateIndexError(
            "The following error occurred trying to create the index [(index_name: " +
            str(index_name) + ") (schema: "+str(schema)+") (definition: "+str(definition) +
            ") (index_type: "+str(index_type)+")]: \n"+str(e))


@db_exceptions_decorator
def search_from_index(index, query_string):
    '''
    Performs a search on the index using the provided query string

    @param index: the index to search on
    @param query_string: the query string to use

    @return: the results of the search

    @raise DatabaseIndexSearchError if an error occurrs during the search operation on the index.
    '''
    try:
        return index.search(query_string)
    except Exception as e:
        raise DatabaseIndexSearchError(
            "The following error occurred trying to search the index [(index_name: " +
            str(index.name) + ") (query_string: "+str(query_string)+")]: \n"+str(e))


@db_exceptions_decorator
def update_object(connection, object_name, to_update_object):
    '''
    Updates the object with the given name from the database using the provided connection.

    @param connection: the redis connection to use
    @param object_name: the name of the object to update
    @param to_update_object: the object to update

    @return: the updated object

    @raise DatabaseUpdateError if an error occurrs during the update operation.
    '''
    try:
        return persist_new_object(connection, object_name, to_update_object)
    except Exception as e:
        raise DatabaseUpdateError(
            "The following error occurred trying to update the object [(object_name: " + object_name +
            ") (to_update_object: "+str(to_update_object)+")]: \n"+str(e))


@db_exceptions_decorator
def delete_object(connection, object_name):
    '''
    Deletes the object with the given name from the database using the provided connection.

    @param connection: the redis connection to use
    @param object_name: the name of the object to delete    

    @return: the number of deleted objects

    @raise DatabaseError if an error occurrs during the delete operation.
    '''
    try:
        return connection.json().delete(object_name)
    except Exception as e:
        raise DatabaseDeleteError(
            "The following error occurred trying to delete the object [(object_name: " + str(object_name) +
            ")]: \n"+str(e))



import logging

from logging.handlers import TimedRotatingFileHandler
from mqtt_middlware_utils.exceptions.mqtt_middleware_exceptions import LoggerInitException, LoggerException


def map_log_level(log_level_string):
    '''
    Maps the provided string to a logging level. 
    If the provided string can't be mapped to a valid logging level, 
    the default level INFO is selected.

    @param log_level_string string to map to

    @return mapped logging level
    '''
    mapped_log_level = logging.INFO
    if log_level_string == "DEBUG":
        mapped_log_level = logging.DEBUG
    if log_level_string == "ERROR":
        mapped_log_level = logging.ERROR
    if log_level_string == "WARNING":
        mapped_log_level = logging.WARNING

    return mapped_log_level


def init_logger(log_level, logger_name, log_destination):
    '''
    Initialize a logger with the provided logger_name.

    @param log_level logging level of the logger
    @param logger_name name of the logger
    @param log_destination path to the log file

    @return the initialized logger instance

    @raise LoggerInitException if an error occurs while initializing the logger

    '''
    try:
        # create custom logger
        logger = logging.getLogger(logger_name)
        # set logging level
        logger.setLevel(map_log_level(log_level))
        # create handler
        # TimedRotatingFileHandler handles log files rotating based on time
        handler = TimedRotatingFileHandler(
            log_destination, "D", utc=True, backupCount=30)
        # set format for logging
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger
    except Exception as e:
        raise LoggerInitException(
            "The following error occurred while initializing the logger: " + str(e))


def log_info(logger_name, msg):
    '''
    Logs the provided message as INFO.

    @param logger_name name of the logger
    @param msg the message to be logged

    @raise LoggerException if an error occurs while logging info

    '''
    try:
        logging.getLogger(logger_name).info(msg)
    except Exception as e:
        raise LoggerException(
            "The following error occurred while logging info: " + str(e))


def log_debug(logger_name, msg):
    '''
    Logs the provided message as DEBUG.

    @param logger_name name of the logger
    @param msg the message to be logged

    @raise LoggerException if an error occurs while logging debug

    '''
    try:
        logging.getLogger(logger_name).debug(msg)
    except Exception as e:
        raise LoggerException(
            "The following error occurred while logging debug: " + str(e))


def log_error(logger_name, msg):
    '''
    Logs the provided message as ERROR.

    @param logger_name name of the logger
    @param msg the message to be logged

    @raise LoggerException if an error occurs while logging error

    '''
    try:
        logging.getLogger(logger_name).error(msg)
    except Exception as e:
        raise LoggerException(
            "The following error occurred while logging error: " + str(e))


def log_warning(logger_name, msg):
    '''
    Logs the provided message as WARNING.

    @param logger_name name of the logger
    @param msg the message to be logged

    @raise LoggerException if an error occurs while logging warning
    '''
    try:
        logging.getLogger(logger_name).warning(msg)
    except Exception as e:
        raise LoggerException(
            "The following error occurred while logging warning: " + str(e))
